""" BitBlock
    Copyright (C) 2022  C. Lockett

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

""" Custom types used throughout BitBlock to match the Bitcoin RPC
return style. These allow for better understanding of code throughout
the module, and a lot less confusion.

At least, for me.

"""

from typing import Any, Dict, List, TypeAlias, TypeVar, Union
from typing_extensions import Literal, TypedDict

BestBlockHash: TypeAlias = str
BlockHash: TypeAlias = str
BlockCount: TypeAlias = str
ConnectionCount: TypeAlias = int
Difficulty: TypeAlias = float
HelpText: TypeAlias = str
NetworkHashPS: TypeAlias = float

class MempoolInfo(TypedDict):
    """ Active state of the TX memory pool. 
    
    ### Properties
    --------------

    loaded: bool
        true if mempool is fully loaded

    size: int
        current tx count

    bytes: int
        sum of all virtual tx sizes

    usage: int
        total memory usage for the mempool

    maxmempool: int
        maximum memory usage for the mempool

    mempoolminfee: float
        minimum fee in BTC/kB for tx to be accepted

    minrelaytxfee: float
        current minimum relay fee for tx

    unbroadcastcount: int
        current number of tx w/o initial broadcast
    
    """

    loaded: bool
    size: int
    bytes: int
    usage: int
    maxmempool: int
    mempoolminfee: float
    minrelaytxfee: float
    unbroadcastcount: int


class _NetworkInfoNetworks(TypedDict):
    """ Information by network. Used by NetworkInfo.
    
    ### Properties
    --------------

    name: Literal['ipv4', 'ipv6', 'onion']
        network name

    limited: bool
        is the network limited using -onlynet?

    reachable: bool
        is the network reachable?

    proxy: str
        the proxy that is used for this network

    proxy_randomize_credentials: bool
        whether randomized credentials are used

    """
    name: Literal['ipv4', 'ipv6', 'onion']
    limited: bool
    reachable: bool
    proxy: str
    proxy_randomize_credentials: bool


class _NetworkInfoLocalAddresses(TypedDict):
    """ List of local addresses. Used by NetworkInfo.

    ### Properties
    --------------
    address: str
        network address

    port: int
        network port

    score: float
        relative score

    """
    address: str
    port: int
    score: float


class NetworkInfo(TypedDict):
    """ State info regarding P2P networking.

    ### Properties
    --------------

    version: int
        the server version

    subversion: str
        the server subversion string

    protocolversion: int
        the protocol version

    localservices: str
        the services we offer to the network in hex

    localservicesname: List[str]
        services we offer to the network

    localrelay: bool
        true if transaction relay is requested from peers

    timeoffset: int
        the time offset

    connections: int
        the total number of connections

    connections_in: int
        the number of inbound connections

    connections_out: int
        the number of outbound connections

    networkactive: bool
        whether p2p networking is enabled

    networks: List[_NetworkInfoNetworks]
        information per network

    relayfee: float
        minimum relay fee for tx in BTC/kB

    incrementalfee: float
        minimum fee increment for mempool limiting

    localaddresses: List[_NetworkInfoLocalAddresses]
        list of local addresses

    warnings: str
        any network and blockchain warnings

    """
    version: int
    subversion: str
    protocolversion: int
    localservices: str
    localservicesnames: List[str]
    localrelay: bool
    timeoffset: int
    connections: int
    connections_in: int
    connections_out: int
    networkactive: bool
    networks: List[_NetworkInfoNetworks]
    relayfee: float
    incrementalfee: float
    localaddresses: List[_NetworkInfoLocalAddresses]
    warnings: str


class _ChainTipsDetail(TypedDict):
    height: int
    hash: str
    branchlen: int
    status: Literal["active", "valid-fork", "valid-headers",
        "headers-only", "invalid"]


ChainTips = List[_ChainTipsDetail]


class _BlockHeader(TypedDict):
    hash: BlockHash
    confirmations: int
    height: int
    version: int
    versionHex: str
    merkleroot: str
    time: int
    mediantime: int
    nonce: int
    bits: str
    difficulty: float
    chainwork: str
    nTx: int
    previousblockhash: str
    nextblockhash: str


BlockHeader = Union[str, _BlockHeader]


class BlockStats(TypedDict):
    avgfee: int
    avgfeerate: int
    avgtxsize: int
    blockhash: str
    feerate_percentiles: List[int]
    height: int
    ins: int
    maxfee: int
    maxfeerate: int
    maxtxsize: int
    medianfee: int
    medintxsize: int
    minfee: int
    minfeerate: int
    mintxsize: int
    outs: int
    subsidy: int
    swtotal_size: int
    swtotal_weight: int
    swtxs: int
    time: int
    total_out: int
    total_size: int
    total_weights: int
    totalfee: int
    txs: int
    utxo_increase: int
    utxo_size_inc: int


class _RawTransactionVIn(TypedDict):
    txid: str
    vout: int
    scriptSig: Dict[str, str]
    sequence: int
    txinwitness: List[str]


class _RawTransactionScriptPubKey(TypedDict):
    asm: str
    hex: str
    reqSigs: int
    type: str
    addresses: List[str]


class _RawTransactionVOut(TypedDict):
    value: float
    n: int
    scriptPubKey: _RawTransactionScriptPubKey


class _RawTransaction(TypedDict):
    txid: str
    hash: str
    version: int
    size: int
    vsize: int
    weight: int
    locktime: int
    vin: List[_RawTransactionVIn]
    vout: List[_RawTransactionVOut]
    blockhash: str
    confirmations: int
    blocktime: int
    time: int


RawTransaction = Union[str, _RawTransaction]

class Block(_BlockHeader):
    strippedsize: int
    tx: List[RawTransaction]


class MiningInfo(TypedDict):
    blocks: int
    currentblockweight: int
    currentblocktx: int
    difficulty: float
    networkhashps: NetworkHashPS
    pooledtx: int
    chain: Literal["main", "test", "regtest"]
    warnings: str

BitcoinRPCResponse = TypeVar(
    "BitcoinRPCResponse",
    BestBlockHash,
    Block,
    BlockCount,
    BlockHash,
    BlockHeader,
    BlockStats,
    ChainTips,
    ConnectionCount,
    Difficulty,
    MempoolInfo,
    MiningInfo,
    NetworkHashPS,
    NetworkInfo,
    RawTransaction,
)

class TransactionDebit(TypedDict):
    txid: str
    n: int
    addresses: List[str]
    value: float

class _TransactionCredit(TypedDict):
    addresses: List[str]
    value: float

TransactionCredit: TypeAlias = List[_TransactionCredit]

class Transactions(TypedDict):
    txid: str
    debit: TransactionDebit
    credits: TransactionCredit


""" BitBlock v0.0.1 """

import asyncio
import sqlite3 as sql3
import threading
import time
from typing import Tuple, Optional

import nest_asyncio
nest_asyncio.apply()

from .rpc._types import *
from . import rpc

# Constants
GENESIS_BLOCK_HASH: BlockHash = "000000000019d6689c085ae165831e934ff763ae46a2a6c172b3f1b60a8ce26f"

# Helper
def async_get(obj):
    _task = asyncio.create_task(obj)
    try:
        asyncio.get_running_loop().run_until_complete(_task)
    except RuntimeError:
        asyncio.new_event_loop().run_until_complete(_task)
    return _task.result()

def print_progress(
    job: str, 
    current: int, total: int, first: int,
    start: float, prev: float
):
    _percent: float = (current / total) * 100.00
    print(
        f"{job} [{'=' * int(_percent // 10)}"
        + f"{' ' * (10 - int(_percent // 10))}]"
        + f" {' ' if _percent < 10 else ''}{_percent:.2f}%"
        + f" ({current}/{total})"
        + f" avg. {(prev - start) / (current - first):.2f}s/block,"
        + f" est. {(total - current) * (prev - start) / (current - first):.2f}s"
        + f" remaining",
        end='\r'
    )

class BtcBlock(object):
    """ A block on the Bitcoin blockchain with a few methods that are easier
    to work with than the initial output from RPC.

    """
    __slots__ = ("_rpc_conn", "_block_hash", "_next_block_hash", 
        "_prev_block_hash", "_transactions", "_height", "_time")
    def __init__(self,
    rpc_conn: rpc.BitcoinRPC, block_hash: BlockHash):
        """ Initialises a BtcBlock. """
        self._rpc_conn: rpc.BitcoinRPC = rpc_conn
        self._block_hash: BlockHash = block_hash
        async_get(self._update_data())
    
    @classmethod
    def from_hash(cls,
        rpc_conn: rpc.BitcoinRPC, block_hash: BlockHash):
        """ Initialises a BtcBlock by its hash. """
        return cls(rpc_conn, block_hash)
    
    @classmethod
    async def from_height(cls,
        rpc_conn: rpc.BitcoinRPC, block_height: int):
        """ Initialises a BtcBlock by its height. """
        _block_hash: BlockHash = await rpc_conn.get_block_hash(block_height)
        return cls(rpc_conn, _block_hash)

    async def _update_data(self) -> None:
        """ Internal.
        
        Updates the information in the BtcBlock from the RPC.

        """
        _block: Block = await self._rpc_conn.get_block(
            block_hash = self._block_hash,
            verbosity = 2
        )
        self._height: int = _block["height"]
        try:
            self._next_block_hash: BlockHash = _block["nextblockhash"]
        except KeyError:
            self._next_block_hash: BlockHash = self._block_hash
        try:
            self._prev_block_hash: BlockHash = _block["previousblockhash"]
        except KeyError:
            self._prev_block_hash: BlockHash = self._block_hash
        self._time: float = _block["time"]
        self._transactions = await self._async_process_transactions(_block["tx"])
    
    async def _async_process_transactions(self,
        tx_list: List[RawTransaction]) -> List:
        """ Internal.

        Asynchroniously handles transactions.
        """
        _all_tx = []
        for _tx in tx_list:
            _tx = asyncio.create_task(self._async_interpret_transaction(_tx))
            _all_tx.append(
                await _tx
            )
        return _all_tx
            
    async def _async_interpret_transaction(self, _tx: RawTransaction):
        _tx_txid: str = _tx["txid"]
        _input = _tx["vin"]
        _tx_input: List = []
        _tx_output: List = []

        # Process the transaction inputs.
        for _i in _input:
            # Check whether or not the input is a coinbase.
            _coinbase = "coinbase" in _input[0].keys()
            # Check if there's an n value in this input.
            _n_exist = "n" in _input[0].keys()
            # Handle a coinbase transaction.
            if _coinbase:
                _tx_in_txid = _i["coinbase"]
                _tx_in_n: int = _i["sequence"]
                _tx_in_addr: List[str] = ["coinbase"]
                _tx_in_value: float = 50.00
            # Handle non-coinbase transaction.
            else:
                _tx_in_txid = _i["txid"]
                _tx_in_n: int = _i["n"] if _n_exist else 0
                # Pull the transaction from which this transaction originates.
                _input_tx = await self._rpc_conn.get_raw_transaction(
                    _tx_in_txid,
                    verbose = True
                )
                _in_vout = _input_tx["vout"][_tx_in_n]
                _in_spk = _in_vout["scriptPubKey"]
                if "addresses" in _in_spk.keys():
                    _tx_in_addr: List[str] = _in_spk["addresses"]
                else:
                    _tx_in_addr: List[str] = [_in_spk["hex"]]
                _tx_in_value: float = _in_vout["value"]
            # Append the input to the main input list.
            _tx_input.append((
                _tx_in_txid,
                _tx_in_n,
                _tx_in_addr,
                _tx_in_value,
            ))
        # Process transaction output
        for _o in _tx["vout"]:
            _out_spk = _o["scriptPubKey"]
            if "addresses" in _out_spk.keys():
                _tx_out_addr: List[str] = _out_spk["addresses"]
            else:
                _tx_out_addr: List[str] = [_out_spk["hex"]]
            _tx_out_value: float = _o["value"]
            _tx_output.append((
                _tx_out_addr,
                _tx_out_value,
            ))
        # Return the transaction.
        return [_tx_txid, _tx_input, _tx_output]
    
    def get_hash(self) -> BlockHash:
        """ Returns the block hash. """
        return self._block_hash
    
    def get_height(self) -> int:
        """ Returns the block height. """
        return self._height
    
    def get_next_hash(self) -> BlockHash:
        """ Returns the hash of the next block in the chain. """
        return self._next_block_hash
    
    def get_previous_hash(self) -> BlockHash:
        """ Returns the hash of the next block in the chain. """
        return self._prev_block_hash
    
    def get_transactions(self) -> List:
        """ Returns the list of transactions in this block. """
        return self._transactions

    def get_time(self) -> float:
        """ Returns the blocktime as UNIX epoch. """
        return self._time


class BitBlockCache(object):
    """ A SQLite3 database to store data from the Bitcoin blockchain with 
    information not readily available from the blockchain distilled for quick
    perusal.

    """
    __slots__ = ("_db_location", "_db", "_cursor")
    def __init__(self, db_location: str):
        self._db_location = db_location
        self._init_db()
    
    def _open(self) -> None:
        """ Internal.

        Connects to the cache database.

        """
        self._db = sql3.connect(self._db_location)
        self._cursor = self._db.cursor()
    
    def _save_and_close(self) -> None:
        """ Internal.

        Saves and closes the connection to the cache database.

        """
        if self._db:
            self._db.commit()
            self._db.close()
            self._cursor = None
            self._db = None
    
    def _close_no_save(self) -> None:
        """ Internal.

        Closes cache database connection without saving changes.

        """
        if self._db:
            self._db.close()
            self._cursor = None
            self._db = None
    
    def _init_db(self) -> None:
        """ Internal

        Checks for needed tables and adds them if not present.

        """
        self._open()
        _i_tx = not self.table_exists("transactions")
        _i_bal = not self.table_exists("balances")
        if _i_tx or _i_bal:
            if _i_tx:
                self._cursor.execute("""
                    CREATE TABLE transactions
                        (block_hash TEXT,
                         block_height NUMERIC,
                         txid TEXT,
                         txtime NUMERIC,
                         debit_txid TEXT,
                         debit_n NUMERIC,
                         debit_addresses TEXT,
                         debit_value NUMERIC,
                         credit_addresses TEXT,
                         credit_value NUMERIC
                        )
                """)
            if _i_bal:
                self._cursor.execute("""
                    CREATE TABLE balances
                        (addresses TEXT,
                         balance NUMERIC,
                         last_used NUMERIC
                        )
                """)
            self._save_and_close()
        return None
    
    def table_exists(self, table_name: str) -> bool:
        """ Returns whether or not `table_name` exists in the cache. """
        try:
            self._cursor.execute(f"SELECT * FROM {table_name}")
        except sql3.OperationalError:
            return False
        return True
    
    def cache_tx(
        self,
        tx: List,
        block_hash: str,
        block_height: int,
        block_time: int,
        ) -> None:
        """ Inserts `tx` into the cache database. """
        _txid: str = tx[0]
        _d_txid: str = tx[1][0][0]
        _d_n: int = tx[1][0][1]
        _d_addresses: str = ','.join(tx[1][0][2])
        _d_value: float = tx[1][0][3]
        _c_addresses: str = ','.join(tx[2][0][0])
        _c_value: float = tx[2][0][1]
        self._cursor.execute(f"""
            INSERT INTO transactions VALUES (
                "{block_hash}", "{block_height}",
                "{_txid}", {block_time},
                "{_d_txid}", {_d_n}, "{_d_addresses}", {_d_value},
                "{_c_addresses}", {_c_value}
            )
        """)
        self._db.commit()
    
    def cache_multi_tx(
        self,
        b_tx: List,
        block_hash: str,
        block_height: int,
        block_time: int
    ) -> None:
        """ Inserts all transactions in `tx` into the cache database. """
        self._open()
        for _t in b_tx:
            self.cache_tx(_t, block_hash, block_height, block_time)
        self._save_and_close()


class BitBlock(object):
    """ A representation of an RPC connection with a local cache of useful
    information derived from the blockchain.

    """
    __slots__ = ("_cache", "_rpc", "_block", "_auto_update", "_update_queue",
        "_threads")
    def __init__(self,
        rpc_username: str,
        rpc_password: str,
        rpc_url: str,
        db_location: str = ".\\.bitblock.cache.db",
        auto_update: bool = True):
        """ Initialises BitBlock """
        self._cache: BitBlockCache = BitBlockCache(db_location)
        self._rpc = rpc.BitcoinRPC(rpc_url, rpc_username, rpc_password)
        self._block = BtcBlock.from_hash(self._rpc, GENESIS_BLOCK_HASH)
        self._auto_update = auto_update
        self._update_queue = []
        self._threads = []
        async_get(self._update())

    async def _async_get_block(
        self,
        block_height) -> BtcBlock:
        """ Internal.

        Asynchroniously pulls block information from the RPC and returns it.

        """
        return async_get(BtcBlock.from_height(self._rpc, block_height))
        

    async def _cache_block_tx(self, block: BtcBlock):
        """ Internal.

        Asynchroniously process block transactions and write them to the cache.
        """
        self._cache.cache_multi_tx(block.get_transactions())

    async def _a_update(self, height: int):
        """ Internal.

        Individual async update processes.

        """
        _block = await self._async_get_block(height)
        self._update_queue.append( _block)
    
    def _process_updates(self):
        """ Internal.

        Clears out the update queue.

        """
        while len(self._update_queue) > 0:
            try:
                _block: BtcBlock = self._update_queue.pop(0)
                self._cache.cache_multi_tx(
                    _block.get_transactions(),
                    _block.get_hash(),
                    _block.get_height(),
                    _block.get_time()
                )
                with open('.bitblock.update', 'w+') as _f:
                    _f.write(str(_block.get_height()))
                    _f.close()
            except Exception:
                pass
        
    async def _threaded_block_pull(self, height: int) -> None:
        _thread = threading.Thread(target=self._a_update, args=(height,))
        _thread.start()
        self._threads.append(_thread)
        if len(self._threads) % 5 == 0:
            for thread in self._threads:
                thread.join()

    async def _update(self) -> None:
        """ Internal.

        Updates BitBlock as needed.

        """
        # Try to open the .bitblock.update and get the last processed block.
        _last_height: int = 0
        with open(".bitblock.update", "a+") as _f:
            _f.seek(0, 0)
            try:
                _last_height = int(_f.readlines()[0])
            except IndexError:
                pass
            _f.close()
        # Pull the current best block height.
        _best_hash: str = async_get(self._rpc.get_best_block_hash())
        _best_height: int = async_get(
            self._rpc.get_block(_best_hash)
        )["height"]
        _start_time = time.time()
        _last_time = time.time()
        _start_height = _last_height
        while _last_height < _best_height:
            await asyncio.create_task(self._a_update(_last_height))
            await asyncio.to_thread(self._process_updates)
            _last_height += 1

            print_progress(
                "pull block",
                _last_height, _best_height, _start_height,
                _start_time, _last_time
            )
            _last_time = time.time()
    
    def update_manual(self) -> None:
        """ Manually forces the cache to update itself. """
        _setting = self._auto_update
        self._auto_update = True
        self._update()
        self._auto_update = _setting


""" BitBlock.RPC """

import itertools
from types import TracebackType
from typing import Any, List, Optional, Type, Union, Dict

import httpx
import orjson
from typing_extensions import Literal

from ._exceptions import ConfigurationError, RPCError
from ._types import (
    BestBlockHash,
    BitcoinRPCResponse,
    Block,
    BlockCount,
    BlockHash,
    BlockHeader,
    BlockStats,
    ChainTips,
    ConnectionCount,
    Difficulty,
    MempoolInfo,
    MiningInfo,
    NetworkHashPS,
    NetworkInfo,
    RawTransaction,
    HelpText
)

_next_rpc_id = itertools.count(1).__next__

class BitcoinRPC:
    """ Connection to the BitcoinRPC with async. """
    __slots__ = ("_url", "_client")

    def __init__(self, url: str, rpc_user: str, rpc_password: str, **options: Any) -> None:
        self._url = url
        self._client = self._configure_client(rpc_user, rpc_password, **options)
    
    async def __aenter__(self) -> "BitcoinRPC":
        return self
    
    async def __aexit__(self, exc_type: Type[BaseException], exc_val: BaseException, exc_tb: TracebackType) -> None:
        await self.aclose()

    @staticmethod
    def _configure_client(rpc_user: str, rpc_password: str, **options: Any) -> httpx.AsyncClient:
        """ Configure `httpx.AsyncClient` """
        auth = (rpc_user, rpc_password)
        headers = {"content-type": "application/json"}
        options = dict(options)
        if not options:
            return httpx.AsyncClient(auth=auth, headers=headers)
        if "auth" in options:
            raise ConfigurationError("`auth` set in options")
        if "headers" in options:
            _additional_headers = dict(options.pop("headers"))
            headers.update(_additional_headers)
            headers["content-type"] = "application/json"
        return httpx.AsyncClient(auth=auth, headers=headers, **options)
    
    @property
    def url(self) -> str:
        return self._url
    
    @property
    def client(self) -> httpx.AsyncClient:
        return self._client
    
    async def aclose(self) -> None:
        await self.client.aclose()
    
    async def acall(self, method: str, params: List[Union[str, int, List[str], None]], **kwargs: Any) -> BitcoinRPCResponse:
        """ Construct a custom async request from the Bitcoin RPC. """
        req = self.client.post(
            url=self.url, 
            content=orjson.dumps(
                {
                    "jsonrpc": "2.0",
                    "id": _next_rpc_id(),
                    "method": method,
                    "params": params,
                }
            ),
            **kwargs
        )
        resp = orjson.loads((await req).content)
        if resp["error"] is not None:
            raise RPCError(resp["error"]["code"], resp["error"]["message"])
        else:
            return resp["result"]
    
    async def get_best_block_hash(self) -> BestBlockHash:
        """ Returns the hash of the best (tip) block in the most-work
        fully-validated chain.
        
        """
        return await self.acall("getbestblockhash", [])
    
    async def get_block(self, block_hash: str, verbosity: Literal[0, 1, 2] = 1, timeout: Optional[float] = 30.0) -> Block:
        """ Return changes based upon verbosity setting. """
        return await self.acall("getblock", [block_hash, verbosity], timeout=httpx.Timeout(timeout))

    async def get_block_count(self) -> BlockCount:
        """ Returns the height of the most-work fully-validated chain. """
        return await self.acall("getblockcount", [])

    async def get_block_filter(self, block_hash: str, filtertype: Optional[str] = None) -> Any:
        # TODO: Implement get_block_filter
        raise NotImplementedError("getblockfilter is not implemented.")
    
    async def get_block_hash(self, height: int) -> BlockHash:
        """ Returns hash of block in best-block-chain at height provided. """
        return await self.acall("getblockhash", [height], timeout=httpx.Timeout(30.0))

    async def get_block_header(self, block_hash: str, verbose: Optional[bool] = True) -> BlockHeader:
        """ Return changes based upon whether verbose is requested. """
        return await self.acall("getblockheader", [block_hash, verbose])

    async def get_block_stats(self, hash_or_height: Union[int, str], *keys: str, timeout: Optional[float] = 30.0) -> BlockStats:
        """ Compute per block statistics for a given window. All
        amounts are in satoshis.
        
        """
        return await self.acall("getblockstats", [hash_or_height, list(keys) or None], timeout=httpx.Timeout(timeout))

    async def get_chain_tips(self) -> ChainTips:
        """ Return information about all known tips in the block tree, 
        including the main chain as well as orphaned branches.
        
        """
        return await self.acall("getchaintips", [])

    async def get_chain_tx_stats(self, nblocks: Optional[int], block_hash: Optional[str]) -> Any:
        # TODO: Implement get_chain_tx_stats
        raise NotImplementedError("getchaintxstats is not implemented.")
    
    async def get_difficulty(self) -> Difficulty:
        """ Returns the proof-of-work difficulty as a multiple of the 
        minimum difficulty.
        
        """
        return await self.acall("getdifficulty", [])
    
    async def get_mempool_ancestors(self, txid: str, verbose: Optional[bool] = False) -> Any:
        # TODO: Implement get_mempool_ancestors
        raise NotImplementedError("getmempoolancestors is not implemented.")
    
    async def get_mempool_descendants(self, txid: str, verbose: Optional[bool] = False) -> Any:
        # TODO: Implement get_mempool_descendants
        raise NotImplementedError("getmempooldescendants is not implemented.")
    
    async def get_mempool_entry(self, txid: str) -> Any:
        # TODO: Implement get_mempool_entry
        raise NotImplementedError("getmempoolentry is not implemented.")
    
    async def get_mempool_info(self) -> MempoolInfo:
        """ Returns details on active state of the TX memory pool. """
        return await self.acall("getmempoolinfo", [])
    
    async def get_raw_mempool(self) -> Any:
        # TODO: Implement get_raw_mempool
        raise NotImplementedError("getrawmempool is not implemented.")

    async def get_tx_out(self, txid: str, n: int, include_mempool: Optional[bool] = True) -> Any:
        """ Returns details about an unspent transaction output. """
        # TODO: Implement get_tx_out
        raise NotImplementedError("gettxout is not implemented.")
    
    async def get_tx_out_proof(self, txid: List[str], block_hash: Optional[str]) -> Any:
        # TODO: Implement get_tx_out_proof
        raise NotImplementedError("gettxoutproof is not implemented.")
    
    async def get_tx_out_set_info(self, hash_type: Optional[Literal["hash_serialized_2", "none"]] = "hash_serialized_2") -> Any:
        # TODO: Implement get_tx_out_set_info
        raise NotImplementedError("gettxoutsetinfo is not implemented.")
    
    async def precious_block(self, block_hash: str) -> None:
        """ Treats a block as if it were received before others with 
        the same work.
        
        """
        await self.acall("preciousblock", [block_hash])
    
    async def prune_blockchain(self, height: int) -> int:
        """ Prunes the blockchain to the provided height. Returns the 
        height of the last block pruned.
        
        """
        return await self.acall("pruneblockchain" [height])
    
    async def save_mempool(self) -> None:
        """ Dumps the mempool to disk. It will fail until the previous 
        dump is fully loaded.
        
        """
        await self.acall("savemempool", [])
    
    async def verify_chain(self, check_level: Optional[Literal[0, 1, 2, 3, 4]] = 3, n_blocks: Optional[int] = 6) -> bool:
        """ Verifies blockchain database. """
        return await self.acall("verifychain", [check_level, n_blocks])
    
    async def verify_tx_out_proof(self, proof: str) -> List[str]:
        """ Verifies that a proof points to a transaction in a block, 
        returning the transaction it commits to and throwing an RPC 
        error if the block is not in our best chain
        
        """
        return await self.acall("verifytxoutproof", [proof])
    
    async def get_memory_info(self, mode: Optional[Literal["stats", "mallocinfo"]] = "stats") -> Any:
        # TODO: Implement get_memory_info
        raise NotImplementedError("getmemoryinfo is not implemented.")
    
    async def get_rpc_info(self) -> Any:
        # TODO: Implement get_rpc_info
        raise NotImplementedError("getrpcinfo is not implemented.")
    
    async def help(self, command: Optional[str] = None) -> HelpText:
        """ List all commands, or get help for a specified command. """
        return await self.acall("help", [command])

    async def logging(self, include: Optional[List[str]] = None, exclude: Optional[List[str]] = None) -> Dict[str, bool]:
        """ Gets and sets the logging configuration. """
        return await self.acall("logging", [list(include) or None, list(exclude) or None])

    async def stop(self) -> Literal["Bitcoin Core stopping"]:
        """ Request a graceful shutdown of Bitcoin Core. """
        return await self.acall("stop", [])
    
    async def uptime(self) -> int:
        """ Returns the total uptime of the server. """
        return await self.acall("uptime", [])
    
    async def generate_block(self, output: str, rawtx_or_txid: List[str]) -> BlockHash:
        """ Mine a block with a set of ordered transactions immediately 
        to a specified address or descriptor (before the RPC call 
        returns)
        
        """
        return await self.acall("generateblock", [])
    
    async def generate_to_address(self, n_blocks: int, address: str, maxtries: Optional[int] = 1000000) -> List[BlockHash]:
        """ Mine blocks immediately to a specified address (before the 
        RPC call returns)
        
        """
        return await self.acall("generatetoaddress", [n_blocks, address, maxtries])
    
    async def generate_to_descriptor(self, num_blocks: int, descriptor: str, maxtries: Optional[int] = 1000000) -> List[BlockHash]:
        """ Mine blocks immediately to a specified descriptor (before 
        the RPC call returns)
        
        """
        return await self.acall("generatetodescriptor", [num_blocks, descriptor, maxtries])
    
    async def get_block_template(self, template_request: Any) -> Any:
        # TODO: Implement get_block_template
        raise NotImplementedError("getblocktemplate is not implented.")

    async def get_mining_info(self) -> MiningInfo:
        """ Returns an object containing mining-related information """
        return await self.acall("getmininginfo", [])
    
    async def get_network_hash_ps(self, nblocks: Optional[int] = 120, height: Optional[int] = -1) -> NetworkHashPS:
        """ Returns the estimated network hashes per second based on 
        the last n blocks.
        
        """
        return await self.acall("getnetworkhashps", [nblocks, height])
    
    async def prioritise_transaction(self, txid: str, fee_delta: int) -> Literal[True]:
        """ Accepts the transaction into mined blocks at a higher (or 
        lower) priority.
        
        """
        return await self.acall("prioritisetransaction", [txid, 0, fee_delta])

    async def get_raw_transaction(self, txid: str, verbose: Optional[bool] = False, block_hash: Optional[str] = None) -> RawTransaction:
        """ Return the raw transaction data. """
        return await self.acall("getrawtransaction", [txid, verbose, block_hash])
    
    async def get_network_info(self) -> NetworkInfo:
        """ Returns an object containing various state info regarding 
        P2P networking.
        
        """
        return await self.acall("getnetworkinfo", [])
    
    async def get_connection_count(self) -> ConnectionCount:
        """ Returns the number of connections to other nodes. """
        return await self.acall("getconnectioncount", [])



""" BitBlock
    Copyright (C) 2022  C. Lockett

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class ConfigurationError(Exception):
    """ Exception raised when the RPCConnection has been configured 
    improperly.

    """
    pass


class RPCError(Exception):
    """ Exception raised when the Bitcoin RPC returns an error. """
    def __init__(self, code: int, message: str) -> None:
        self.code = code
        self.message = message